//css_ref Produmex.Foundation.SlimScreen.Interfaces;
//css_ref Produmex.Foundation.Wwf.Sbo;
//css_ref Produmex.Foundation.Data.dll;
//css_ref Produmex.Foundation.Data.Sbo.dll;
//css_ref Produmex.Foundation.dll;
//css_ref Produmex.Foundation.Sbo.dll;
//css_ref Produmex.Sbo.Logex.Data.dll;
//css_ref Produmex.Sbo.Logex.LenL.dll;


using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Globalization;
using System.Reflection;
using System.Text;
using Produmex.Foundation;
using Produmex.Foundation.Data.Sbo;
using Produmex.Foundation.Data.Sbo.BusinessObjects;
using Produmex.Foundation.Data.Sbo.BusinessObjects.Definitions;
using Produmex.Foundation.Data.Sbo.BusinessObjects.Definitions.Tables;
using Produmex.Foundation.Data.Sbo.DataObjects;
using Produmex.Foundation.Data.Sbo.Providers;
using Produmex.Foundation.Data.Sbo.Utilities;
using Produmex.Foundation.Data.SqlClient;
using Produmex.Foundation.Diagnostics;
using Produmex.Foundation.Messages;
using Produmex.Foundation.SlimScreen;
using Produmex.Foundation.Workflows;
using Produmex.Foundation.Workflows.Parameters;
using Produmex.Foundation.Wwf.Sbo.LocalServices;
using Produmex.Sbo.Logex.Data.BusinessObjects;
using Produmex.Sbo.Logex.Data.BusinessObjects.Definitions.Tables;
using Produmex.Sbo.Logex.Data.DataObjects;
using Produmex.Sbo.Logex.Data.Providers;
using Produmex.Sbo.Logex.LenL;
using Produmex.Sbo.Logex.LenL.BusinessObjects;
using Produmex.Sbo.Logex.LenL.BusinessObjects.Definitions;
using Produmex.Sbo.Logex.LenL.BusinessObjects.Definitions.Tables;
using Produmex.Sbo.Logex.LenL.Providers;

namespace WorkflowScript_SalesDeliveryCreatedHookScript
{
    /// <summary>
    /// This is a hook script, meant to be customized on a "per project" basis
    /// </summary>
    public class WorkflowScript_SalesDeliveryCreatedHookScript : WorkflowInstanceScriptBase
    {
        private static readonly ILog s_log = LogProvider.GetLogger(MethodInfo.GetCurrentMethod().DeclaringType);

        // Input parameters * do not change *
        public ReadOnlyBinder<CultureInfo> DefaultCultureInfo;
        public ReadOnlyBinder<PmxOseCompany> PmxOseCompany;
        public ReadOnlyBinder<DocumentRef[]> DocumentRefs;

        // Output parameters
        // <none>

        // Sub-flows
        // <none>

        // Private data
        private static CultureInfo upsCulture = (CultureInfo)CultureInfo.GetCultureInfo("nl-NL").Clone();

        private struct Defs
        {
            internal const string CONTACT_NAME = SboContactPersonTable.NAME + "_" + SboContactPersonTable.Columns.Name.NAME;

            internal const string FIRST_DOC_ENTRY = "firstDocEntry";

            internal const int REPORT_CODE = 10;
            internal const int NUMBER_OF_REPORTS = 1;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="WorkflowScript_SalesDeliveryCreatedHookScript"/> class.
        /// </summary>
        /// <param name="parent">The parent.</param>
        /// <param name="factory">The factory.</param>
        public WorkflowScript_SalesDeliveryCreatedHookScript(WorkflowInstanceBase parent, WorkflowInstanceFactory factory)
            : base(parent, factory)
        {
        }

        #region WorkflowInstanceScriptBase Members

        protected override void Execute()
        {
            // Parameters in scope
            Session session = GetScopeParameter("Session") as Session;
            ISboProviderService sboProviderService = GetScopeParameter("<WwfService>ISboService") as ISboProviderService;

            //First print sales delivery document
            try
            {
                if (DocumentRefs != null)
                {
                    foreach (DocumentRef docRef in DocumentRefs.Get())
                    {
                        //print label
                        sboProviderService.InvokeMethodWithDbConnection<PmxOsePrinter>(false, false, null, null,
                           delegate(PmxDbConnection conn, object[] parameters)
                           {
                               PmxPrintReportProvider prov = new PmxPrintReportProvider(conn);

                               Collection<ReportParameter> reportParameters = new Collection<ReportParameter>();
                               reportParameters.Add(new ReportParameter(Produmex.Sbo.Logex.Data.Reports.Definitions.ReportTypeSalesDeliveryNoteDefinition.Parameters.SalesDeliveryNoteDocEntry.NAME, docRef.DocEntry));

                               return prov.PrintReport(reportParameters, Defs.REPORT_CODE, Defs.NUMBER_OF_REPORTS, session.DeviceInstanceId, null, true);

                           });
                    }
                }
            }
            catch (Exception ex)
            {
                //only log a failed print
                s_log.Error("Print sales delivery document", ex);
            }

            string initialErrorKey = null;
            int numberOfLUID = 0;

            //set fixed decimal signs
            upsCulture.NumberFormat.NumberDecimalSeparator = ",";
            upsCulture.NumberFormat.NumberGroupSeparator = "";

        Step_EnterQuantity:

            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IEnterQuantityScreen),
                DefaultCultureInfo.Get(), BuildParamCollection(
                    "InitialErrorKey", initialErrorKey,
                    "TitleKey", "Vul het aantal logistieke eenheden in",
                    "InitialQuantity", 1,
                    "MaximumQuantity", -1
                    ));
            Message msg = WaitForMessage();

            initialErrorKey = null;
            if (!msg.Name.EndsWith(".QuantityEntered"))
            {
                initialErrorKey = "MSG_ERROR_OPERATION_NOT_SUPPORTED";
                goto Step_EnterQuantity;
            }

            numberOfLUID = Convert.ToInt32(ExtractParameter<double>(msg.Parameters, "quantity"));



            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                DefaultCultureInfo.Get(), BuildParamCollection(
                    "InitialErrorKey", null,
                    "MessageKey", "Processing..."));

            string msgKey = null;
            ProdumexError pmxError = null;
            try
            {
                sboProviderService.InvokeMethodWithDbConnection<bool>(true, false, null, null, CreateXMLFile, session.DeviceInstanceId, numberOfLUID);
            }
            catch (ProdumexException ex)
            {
                s_log.Error("", ex);
                msgKey = ex.Message;
                pmxError = ex.Error;
            }
            catch (Exception ex)
            {
                s_log.Error("", ex);
                msgKey = ex.Message;
            }

            if (msgKey != null)
            {
                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                    DefaultCultureInfo.Get(), BuildParamCollection(
                        "InitialErrorKey", null,
                        "MessageKey", msgKey,
                        "PmxError", pmxError,
                        "ShowButton", true));
                WaitForMessage();
            }
        }

        #endregion

        /// <summary>
        /// Creates the XML file.
        /// </summary>
        /// <param name="dbConn">The db conn.</param>
        /// <param name="parameters">The parameters.</param>
        /// <returns></returns>
        private bool CreateXMLFile(PmxDbConnection dbConn, object[] parameters)
        {
            string deviceInstanceId = (string)parameters[0];

            int numberOfLUID = (int)parameters[1];

            PmxParameterValueProvider parProv = new PmxParameterValueProvider(dbConn);
            Dictionary<string, PmxParameterValue> upsParameters = parProv.GetParameterValuesAsDictionary(String.Format("{0}->{1}", deviceInstanceId, PmxPropertyCodes.OSE_PARAMETERSET), true, false);

            PmxSalesDeliveryNoteProvider delNoteProv = new PmxSalesDeliveryNoteProvider((SboConnection)dbConn);


            const string ns = "x-schema:OpenShipments.xdr";

            StringBuilder docRefsStr = new StringBuilder();


            System.Xml.XmlDocument xmlDoc = new System.Xml.XmlDocument();
            System.Xml.XmlElement openShipments = xmlDoc.CreateElement("OpenShipments", ns);
            {
                foreach (DocumentRef docRef in DocumentRefs.Get())
                {
                    string firstTrackingNumber = null;

                    docRefsStr.AppendFormat(", DocType {0} DocEntry {1}", docRef.DocType, docRef.DocEntry);

                    string query = BuildQuery.GetSboSalesDeliveryHeader(docRef.DocEntry, dbConn.DbCultureInfo);
                    using (ISboRecordset rsOrder = SboRecordsetHelper.RunQuery(s_log, query, dbConn))
                    {
                        if (rsOrder.EoF) continue;  // Skip if not UPS or not found

                        System.Xml.XmlElement openShipment = xmlDoc.CreateElement("OpenShipment", ns);
                        {
                            openShipment.SetAttribute("ShipmentOption", "");
                            openShipment.SetAttribute("ProcessStatus", "");

                            // Add ShipTo
                            System.Xml.XmlElement shipTo = xmlDoc.CreateElement("ShipTo", ns);
                            {
                                System.Xml.XmlElement elm = xmlDoc.CreateElement("CompanyOrName", ns);
                                elm.AppendChild(xmlDoc.CreateTextNode(Convert.ToString(rsOrder[SboSalesDeliveryNoteTable.Columns.CardName.NAME])));
                                shipTo.AppendChild(elm);

                                elm = xmlDoc.CreateElement("Attention", ns);
                                elm.AppendChild(xmlDoc.CreateTextNode(Convert.ToString(rsOrder[Defs.CONTACT_NAME])));
                                shipTo.AppendChild(elm);

                                elm = xmlDoc.CreateElement("Address1", ns);
                                elm.AppendChild(xmlDoc.CreateTextNode((Convert.ToString(rsOrder[SboSalesDeliveryTaxExtensionTable.Columns.StreetS.NAME]) + "  " + Convert.ToString(rsOrder["StreetNoS"])).Trim()));
                                shipTo.AppendChild(elm);

                                //add addr2
                                elm = xmlDoc.CreateElement("Address2", ns);
                                elm.AppendChild(xmlDoc.CreateTextNode((Convert.ToString(rsOrder[SboSalesOrderTaxExtensionTable.Columns.Address2S.NAME]))));
                                shipTo.AppendChild(elm);

                                string country = Convert.ToString(rsOrder[SboSalesDeliveryTaxExtensionTable.Columns.CountryS.NAME]);

                                elm = xmlDoc.CreateElement("CountryTerritory", ns);
                                elm.AppendChild(xmlDoc.CreateTextNode(country));
                                shipTo.AppendChild(elm);

                                elm = xmlDoc.CreateElement("PostalCode", ns);
                                elm.AppendChild(xmlDoc.CreateTextNode(Convert.ToString(rsOrder[SboSalesDeliveryTaxExtensionTable.Columns.ZipCodeS.NAME])));
                                shipTo.AppendChild(elm);

                                elm = xmlDoc.CreateElement("CityOrTown", ns);
                                elm.AppendChild(xmlDoc.CreateTextNode(Convert.ToString(rsOrder[SboSalesDeliveryTaxExtensionTable.Columns.CityS.NAME])));
                                shipTo.AppendChild(elm);

                                // For US add state
                                if (String.Compare(country, "US", true) == 0)
                                {
                                    elm = xmlDoc.CreateElement("StateProvinceCounty", ns);
                                    elm.AppendChild(xmlDoc.CreateTextNode(Convert.ToString(rsOrder[SboSalesDeliveryTaxExtensionTable.Columns.StateS.NAME])));
                                    shipTo.AppendChild(elm);
                                }

                                string phoneNr = Convert.ToString(rsOrder[SboBusinessPartnerTable.Columns.Phone1.NAME]);
                                if (String.IsNullOrEmpty(phoneNr)) phoneNr = ".";
                                elm = xmlDoc.CreateElement("Telephone", ns);
                                elm.AppendChild(xmlDoc.CreateTextNode(phoneNr));
                                shipTo.AppendChild(elm);
                            }
                            openShipment.AppendChild(shipTo);

                            // Determine COD
                            double codAmount = 0.0;
                            string codCurrency = null;

                            if (Convert.ToString(rsOrder[SboPaymentTermsTypeTableLenL.Columns.Rembours.NAME]) == SboBooleanColumnValues.Y.VALUE)
                            {
                                codAmount = rsOrder.GetTypedValue<double>(SboSalesDeliveryNoteTable.Columns.DocTotal.NAME);
                                codCurrency = Convert.ToString(rsOrder[SboSalesDeliveryNoteTable.Columns.DocCurrency.NAME]);

                                // If foreign currency, use that amount
                                double codAmountFC = rsOrder.GetTypedValue<double>(SboSalesDeliveryNoteTable.Columns.DocTotalFC.NAME);
                                if (codAmountFC <= -0.005 || codAmountFC >= 0.005)
                                {
                                    codAmount = codAmountFC;
                                }
                            }

                            //if ( Convert.ToString( rsOrder[ SboPaymentTermsTypeTableLenL.Columns.Rembours.NAME ] ) == SboBooleanColumnValues.Y.VALUE )
                            //{
                            //    query = BuildQuery.GetOrderDeliveryInfo( docRef.DocEntry, dbConn.DbCultureInfo );
                            //    using ( ISboRecordset rsOrderDeliveryInfo = SboRecordsetHelper.RunQuery( s_log, query, dbConn ) )
                            //    {
                            //        if ( Convert.ToInt32( rsOrderDeliveryInfo[ Defs.FIRST_DOC_ENTRY ] ) == docRef.DocEntry )
                            //        {
                            //            codAmount = rsOrderDeliveryInfo.GetTypedValue<double>(SboSalesOrderTable.Columns.DocTotal.NAME);
                            //            codCurrency = Convert.ToString(rsOrderDeliveryInfo[SboSalesOrderTable.Columns.DocCurrency.NAME]);

                            //            // If foreign currency, use that amount
                            //            double codAmountFC = rsOrderDeliveryInfo.GetTypedValue<double>(SboSalesOrderTable.Columns.DocTotalFC.NAME);
                            //            if ( codAmountFC <= -0.005 || codAmountFC >= 0.005 )
                            //            {
                            //                codAmount = codAmountFC;
                            //            }
                            //        }
                            //    }
                            //}

                            PmxUPSTrackingProvider upsProv = new PmxUPSTrackingProvider(dbConn);




                            //query = BuildQuery.GetAllLUIDsForSalesDelivery( docRef.DocEntry, dbConn.DbCultureInfo );
                            //using ( ISboRecordset rsLogUnits = SboRecordsetHelper.RunQuery( s_log, query, dbConn ) )
                            {
                                // Add ShipmentInformation
                                System.Xml.XmlElement shipmentInformation = xmlDoc.CreateElement("ShipmentInformation", ns);
                                {
                                    string[] serviceCodes = PmxParameterValueProvider.GetParameterValue<string>(upsParameters, PmxExtensionParameterCodesLenL.UPS_SERVICE_CODE).Split(',', ';', '|', ':');

                                    System.Xml.XmlElement elm = xmlDoc.CreateElement("ServiceType", ns);
                                    elm.AppendChild(xmlDoc.CreateTextNode(serviceCodes[0]));
                                    shipmentInformation.AppendChild(elm);

                                    elm = xmlDoc.CreateElement("NumberOfPackages", ns);
                                    //elm.AppendChild( xmlDoc.CreateTextNode( rsLogUnits.RecordCount.ToString() ) );
                                    elm.AppendChild(xmlDoc.CreateTextNode(numberOfLUID.ToString()));
                                    shipmentInformation.AppendChild(elm);

                                    elm = xmlDoc.CreateElement("BillingOption", ns);
                                    elm.AppendChild(xmlDoc.CreateTextNode("PP"));
                                    shipmentInformation.AppendChild(elm);

                                    // Results in UPS import error
                                    //elm = xmlDoc.CreateElement( "Reference1", ns );
                                    //elm.AppendChild( xmlDoc.CreateTextNode( Convert.ToString( rsOrder[ SboSalesDeliveryNoteTable.Columns.NumAtCard.NAME ] ) ) );
                                    //shipmentInformation.AppendChild( elm );

                                    if (codAmount >= 0.005)
                                    {
                                        System.Xml.XmlElement cod = xmlDoc.CreateElement("COD", ns);
                                        {
                                            elm = xmlDoc.CreateElement("CashOnly", ns);
                                            elm.AppendChild(xmlDoc.CreateTextNode("y"));
                                            cod.AppendChild(elm);

                                            s_log.WarnFormat("The codAmount is: '{0}' AND converted to string: '{1}'", codAmount, codAmount.ToString("F2", upsCulture));

                                            elm = xmlDoc.CreateElement("Amount", ns);
                                            elm.AppendChild(xmlDoc.CreateTextNode(codAmount.ToString("F2", upsCulture)));
                                            cod.AppendChild(elm);

                                            elm = xmlDoc.CreateElement("Currency", ns);
                                            elm.AppendChild(xmlDoc.CreateTextNode(codCurrency));
                                            cod.AppendChild(elm);
                                        }
                                        shipmentInformation.AppendChild(cod);
                                    }
                                }
                                openShipment.AppendChild(shipmentInformation);

                                // Add packages
                                bool codServiceLevel = (codAmount >= 0.005) ? true : false;
                                for (int iLoop = 0; iLoop < numberOfLUID; iLoop++)
                                //while ( !rsLogUnits.EoF )
                                {
                                    string trackingNumber = CreateTrackingNumber(upsParameters, codServiceLevel);

                                    if (string.IsNullOrEmpty(firstTrackingNumber))
                                    {
                                        firstTrackingNumber = trackingNumber;

                                        using (PmxSalesDeliveryNote delNote = delNoteProv.GetBO(docRef.DocEntry))
                                        {
                                            if (delNote != null)
                                            {
                                                //  delNote.TrackingNumber = trackingNumber;

                                                //delNote["TrackNo"] = trackingNumber;

                                                delNote[PmxSalesDeliveryNoteTableLenL.Columns.NumberOfLabels.NAME] = numberOfLUID.ToString();
                                                delNote[PmxSalesDeliveryNoteTableLenL.Columns.DeliveryDevice.NAME] = deviceInstanceId;

                                                delNoteProv.UpdateBO(delNote, false, false);
                                            }
                                        }


                                    }

                                    codServiceLevel = false;  // First package has different service level

                                    //query = BuildQuery.UpdateLUID( rsLogUnits.GetTypedValue<int>( PmxLogisticUnitIDTable.Columns.InternalKey.NAME ), trackingNumber, dbConn.DbCultureInfo );
                                    //SboRecordsetHelper.RunQueryNoReturn( s_log, query, dbConn );

                                    using (PmxUPSTracking tracking = upsProv.GetNewBO())
                                    {
                                        tracking.UPSTrackingNumber = trackingNumber;
                                        tracking.SalesDeliveryDocEntry = docRef.DocEntry;

                                        upsProv.AddBO(tracking);
                                    }

                                    System.Xml.XmlElement package = xmlDoc.CreateElement("Package", ns);
                                    {
                                        System.Xml.XmlElement elm = xmlDoc.CreateElement("PackageType", ns);
                                        elm.AppendChild(xmlDoc.CreateTextNode("CP"));
                                        package.AppendChild(elm);

                                        elm = xmlDoc.CreateElement("Weight", ns);
                                        elm.AppendChild(xmlDoc.CreateTextNode("4"));
                                        package.AppendChild(elm);

                                        elm = xmlDoc.CreateElement("TrackingNumber", ns);
                                        elm.AppendChild(xmlDoc.CreateTextNode(trackingNumber));
                                        package.AppendChild(elm);

                                        elm = xmlDoc.CreateElement("Reference1", ns);
                                        elm.AppendChild(xmlDoc.CreateTextNode(Convert.ToString(rsOrder[SboSalesDeliveryNoteTable.Columns.NumAtCard.NAME])));
                                        package.AppendChild(elm);

                                        elm = xmlDoc.CreateElement("Reference2", ns);
                                        elm.AppendChild(xmlDoc.CreateTextNode(Convert.ToString(rsOrder[SboSalesDeliveryNoteTable.Columns.DocNum.NAME])));
                                        package.AppendChild(elm);
                                    }
                                    openShipment.AppendChild(package);

                                    //rsLogUnits.MoveNext();
                                }
                            }
                        }
                        openShipments.AppendChild(openShipment);
                    }
                }
            }
            xmlDoc.AppendChild(openShipments);

            /* Update TBA: we don't need to write the XML
              if (openShipments.ChildNodes.Count > 0)
              {
                  docRefsStr.Remove(0, 2);
                  s_log.DebugFormat("UPS shipping for one or more of {0}", docRefsStr.ToString());

                  // Update UPS_NEXT_PTN in the database (before writing file so no duplicate PTN's can occur)
                  parProv.UpdateBO(upsParameters[PmxExtensionParameterCodesLenL.UPS_NEXT_PTN], false, false);

                  // Write XML file with encoding "Windows-1252"
                  string outputDir = PmxParameterValueProvider.GetParameterValue<string>(upsParameters, PmxExtensionParameterCodesLenL.UPS_XML_OUT_DIR);
                  System.Xml.XmlWriterSettings settings = new System.Xml.XmlWriterSettings();
                  settings.Encoding = Encoding.GetEncoding(1252);  // This is "Windows-1252"
                  settings.Indent = true;
                  settings.IndentChars = "    ";
                  settings.NewLineChars = "\r\n";
                  using (System.Xml.XmlWriter xmlWriter = System.Xml.XmlWriter.Create(System.IO.Path.Combine(outputDir, "UPSShipmentFile.xml.new"), settings))
                  {
                      xmlDoc.Save(xmlWriter);
                  }
                  try
                  {
                      System.IO.File.Delete(System.IO.Path.Combine(outputDir, "UPSShipmentFile.xml"));
                  }
                  catch
                  {
                      // Throws if not exists, do not log
                  }
                  System.IO.File.Move(System.IO.Path.Combine(outputDir, "UPSShipmentFile.xml.new"), System.IO.Path.Combine(outputDir, "UPSShipmentFile.xml"));


              }
              else
              {
                  docRefsStr.Remove(0, 2);
                  s_log.DebugFormat("No Sales Deliveries found or no UPS shipping for {0}", docRefsStr.ToString());
              }
              */
            return true;
        }

        /// <summary>
        /// Creates the tracking number.
        /// </summary>
        /// <param name="parameters">The parameters.</param>
        /// <param name="collectOnDelivery">if set to <c>true</c> [collect on delivery].</param>
        /// <returns></returns>
        private string CreateTrackingNumber(IDictionary<string, PmxParameterValue> parameters, bool collectOnDelivery)
        {
            string accountNumber = PmxParameterValueProvider.GetParameterValue<string>(parameters, PmxExtensionParameterCodesLenL.UPS_ACCOUNT_NUMBER);
            string[] serviceCodes = PmxParameterValueProvider.GetParameterValue<string>(parameters, PmxExtensionParameterCodesLenL.UPS_SERVICE_CODE).Split(',', ';', '|', ':');
            int firstPTN = PmxParameterValueProvider.GetParameterValue<int>(parameters, PmxExtensionParameterCodesLenL.UPS_FIRST_PTN);
            int lastPTN = PmxParameterValueProvider.GetParameterValue<int>(parameters, PmxExtensionParameterCodesLenL.UPS_LAST_PTN);

            // Use ParameterValue for nextPTN or otherwise we will get null after updating it
            int nextPTN = Convert.ToInt32(parameters[PmxExtensionParameterCodesLenL.UPS_NEXT_PTN].ParameterValue);

            // Build tracking number
            accountNumber = (accountNumber + "000000").Substring(0, 6);  // Fix length to 6
            string serviceIndicator;
            if (collectOnDelivery && serviceCodes.Length >= 3)
            {
                serviceIndicator = (serviceCodes[2].Trim() + "00").Substring(0, 2);  // Fix length to 2
            }
            else
            {
                serviceIndicator = (serviceCodes[1].Trim() + "00").Substring(0, 2);
            }

            string trackingNumber = string.Format("{0}{1}{2:D7}", accountNumber, serviceIndicator, nextPTN);
            int checkDigit = 0;
            for (int i = 0; i < trackingNumber.Length; i++)
            {
                int mul = (i & 0x01) + 1;  // Multiply by 2 for even positions
                if (Char.IsDigit(trackingNumber, i))
                {
                    checkDigit += mul * (trackingNumber[i] - '0');
                }
                else if (trackingNumber[i] >= 'A' && trackingNumber[i] <= 'Z')
                {
                    // A=2, B=3, ...
                    checkDigit += mul * (trackingNumber[i] - 'A' + 2);
                }
            }
            checkDigit = (1000000000 - checkDigit) % 10;

            // Increment PTN
            if (++nextPTN > lastPTN) nextPTN = lastPTN;
            parameters[PmxExtensionParameterCodesLenL.UPS_NEXT_PTN].ParameterValue = nextPTN.ToString();

            // Return tracking number with '1Z' prefix and checkDigit
            return String.Format("1Z{0}{1}", trackingNumber, checkDigit);
        }

        private class BuildQuery
        {
            /// <summary>
            /// Gets the sbo sales delivery header.
            /// </summary>
            /// <param name="docEntry">The doc entry.</param>
            /// <param name="cultInfo">The cult info.</param>
            /// <returns></returns>
            internal static string GetSboSalesDeliveryHeader(int docEntry, CultureInfo cultInfo)
            {
                StringBuilder strBuilder = new StringBuilder();

                strBuilder.AppendFormat("SELECT {0}",
                    SqlCommandHelper.EscapeTableColumnName(SboSalesDeliveryNoteTable.NAME, SboSalesDeliveryNoteTable.Columns.DocNum.NAME));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(SboSalesDeliveryNoteTable.NAME, SboSalesDeliveryNoteTable.Columns.CardName.NAME));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(SboSalesDeliveryNoteTable.NAME, SboSalesDeliveryNoteTable.Columns.NumAtCard.NAME));
                strBuilder.AppendFormat(", {0}",
                   SqlCommandHelper.EscapeTableColumnNameWithAlias(SboSalesDeliveryNoteTable.NAME, SboSalesDeliveryNoteTable.Columns.ShipToCode.NAME, Defs.CONTACT_NAME));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(SboContactPersonTable.NAME, SboContactPersonTable.Columns.Name.NAME));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(SboSalesDeliveryTaxExtensionTable.NAME, SboSalesDeliveryTaxExtensionTable.Columns.StreetS.NAME));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(SboSalesDeliveryTaxExtensionTable.NAME, "StreetNoS"));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(SboSalesDeliveryTaxExtensionTable.NAME, SboSalesDeliveryTaxExtensionTable.Columns.ZipCodeS.NAME));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(SboSalesDeliveryTaxExtensionTable.NAME, SboSalesDeliveryTaxExtensionTable.Columns.CityS.NAME));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(SboSalesDeliveryTaxExtensionTable.NAME, SboSalesDeliveryTaxExtensionTable.Columns.StateS.NAME));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(SboSalesDeliveryTaxExtensionTable.NAME, SboSalesDeliveryTaxExtensionTable.Columns.CountryS.NAME));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(SboSalesDeliveryTaxExtensionTable.NAME, SboSalesDeliveryTaxExtensionTable.Columns.Address2S.NAME));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(SboBusinessPartnerTable.NAME, SboBusinessPartnerTable.Columns.Phone1.NAME));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(SboSalesDeliveryNoteTable.NAME, SboSalesDeliveryNoteTable.Columns.DocCurrency.NAME));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(SboSalesDeliveryNoteTable.NAME, SboSalesDeliveryNoteTable.Columns.DocTotal.NAME));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(SboPaymentTermsTypeTableLenL.NAME, SboPaymentTermsTypeTableLenL.Columns.Rembours.NAME));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(SboSalesDeliveryNoteTable.NAME, SboSalesDeliveryNoteTable.Columns.DocTotalFC.NAME));

                strBuilder.AppendFormat("\r\nFROM {0}",
                    SqlCommandHelper.EscapeTableName(SboSalesDeliveryNoteTable.NAME));

                strBuilder.AppendFormat("\r\nINNER JOIN {0}",
                    SqlCommandHelper.EscapeTableName(SboSalesDeliveryTaxExtensionTable.NAME));
                strBuilder.AppendFormat(" ON {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName(SboSalesDeliveryTaxExtensionTable.NAME, SboSalesDeliveryTaxExtensionTable.Columns.DocEntry.NAME),
                    SqlCommandHelper.EscapeTableColumnName(SboSalesDeliveryNoteTable.NAME, SboSalesDeliveryNoteTable.Columns.DocEntry.NAME));

                strBuilder.AppendFormat("\r\nINNER JOIN {0}",
                    SqlCommandHelper.EscapeTableName(SboBusinessPartnerTable.NAME));
                strBuilder.AppendFormat(" ON {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName(SboBusinessPartnerTable.NAME, SboBusinessPartnerTable.Columns.CardCode.NAME),
                    SqlCommandHelper.EscapeTableColumnName(SboSalesDeliveryNoteTable.NAME, SboSalesDeliveryNoteTable.Columns.CardCode.NAME));

                strBuilder.AppendFormat("\r\nLEFT JOIN {0}",
                    SqlCommandHelper.EscapeTableName(SboPaymentTermsTypeTable.NAME));
                strBuilder.AppendFormat(" ON {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName(SboPaymentTermsTypeTable.NAME, SboPaymentTermsTypeTable.Columns.GroupNum.NAME),
                    SqlCommandHelper.EscapeTableColumnName(SboSalesDeliveryNoteTable.NAME, SboSalesDeliveryNoteTable.Columns.PaymentTermsGroupNum.NAME));

                strBuilder.AppendFormat("\r\nLEFT JOIN {0}",
                    SqlCommandHelper.EscapeTableName(SboContactPersonTable.NAME));
                strBuilder.AppendFormat(" ON {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName(SboContactPersonTable.NAME, SboContactPersonTable.Columns.CntctCode.NAME),
                    SqlCommandHelper.EscapeTableColumnName(SboSalesDeliveryNoteTable.NAME, SboSalesDeliveryNoteTable.Columns.CntctCode.NAME));

                strBuilder.AppendFormat("\r\nWHERE {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName(SboSalesDeliveryNoteTable.NAME, SboSalesDeliveryNoteTable.Columns.DocEntry.NAME),
                    SqlCommandHelper.FormatAndEscapeColumnValue(docEntry, cultInfo));
                strBuilder.AppendFormat(" AND {0} IN (",
                    SqlCommandHelper.EscapeTableColumnName(SboSalesDeliveryNoteTable.NAME, SboSalesDeliveryNoteTable.Columns.TrnspCode.NAME));
                {
                    strBuilder.AppendFormat("    SELECT {0}",
                        SqlCommandHelper.EscapeTableColumnName(SboShippingTypeTable.NAME, SboShippingTypeTable.Columns.Code.NAME));
                    strBuilder.AppendFormat("\r\n    FROM {0}",
                        SqlCommandHelper.EscapeTableName(SboShippingTypeTable.NAME));
                    strBuilder.AppendFormat("\r\n    WHERE {0} LIKE 'UPS%'",
                        SqlCommandHelper.EscapeTableColumnName(SboShippingTypeTable.NAME, SboShippingTypeTable.Columns.Name.NAME));
                    strBuilder.AppendFormat("\r\n    OR {0} LIKE '%UPS'",
                        SqlCommandHelper.EscapeTableColumnName(SboShippingTypeTable.NAME, SboShippingTypeTable.Columns.Name.NAME));
                    strBuilder.AppendFormat("\r\n    OR {0} LIKE '%UPS%'",
                        SqlCommandHelper.EscapeTableColumnName(SboShippingTypeTable.NAME, SboShippingTypeTable.Columns.Name.NAME));
                }
                strBuilder.Append(" )");

                return strBuilder.ToString();
            }

            /// <summary>
            /// Gets the order delivery info.
            /// </summary>
            /// <param name="salesOrderDocEntry">The sales order doc entry.</param>
            /// <param name="dbCulture">The db culture.</param>
            /// <returns></returns>
            internal static string GetOrderDeliveryInfo(int salesDeliveryDocEntry, CultureInfo dbCulture)
            {
                StringBuilder strBuilder = new StringBuilder();

                strBuilder.AppendFormat("SELECT {0}",
                    SqlCommandHelper.EscapeTableColumnName(SboSalesOrderTable.NAME, SboSalesOrderTable.Columns.DocTotal.NAME));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(SboSalesOrderTable.NAME, SboSalesOrderTable.Columns.DocTotalFC.NAME));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(SboSalesOrderTable.NAME, SboSalesOrderTable.Columns.DocCurrency.NAME));
                strBuilder.AppendFormat(", MIN( {0} ) AS {1}",
                    SqlCommandHelper.EscapeTableColumnName(SboSalesDeliveryNoteLineTable.NAME, SboSalesDeliveryNoteLineTable.Columns.DocEntry.NAME),
                    SqlCommandHelper.EscapeColumnName(Defs.FIRST_DOC_ENTRY));

                strBuilder.AppendFormat("\r\nFROM {0}",
                    SqlCommandHelper.EscapeTableName(SboSalesOrderTable.NAME));

                strBuilder.AppendFormat("\r\nLEFT JOIN {0}",
                    SqlCommandHelper.EscapeTableName(SboSalesDeliveryNoteLineTable.NAME));
                strBuilder.AppendFormat(" ON {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName(SboSalesDeliveryNoteLineTable.NAME, SboDocumentLineColumns.BaseType.NAME),
                    SqlCommandHelper.FormatAndEscapeColumnValue(((int)SboObjectType.SalesOrder).ToString(), dbCulture));
                strBuilder.AppendFormat(" AND {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName(SboSalesDeliveryNoteLineTable.NAME, SboDocumentLineColumns.BaseEntry.NAME),
                    SqlCommandHelper.EscapeTableColumnName(SboSalesOrderTable.NAME, SboSalesOrderTable.Columns.DocEntry.NAME));

                strBuilder.AppendFormat("\r\nWHERE {0} IN (",
                    SqlCommandHelper.EscapeTableColumnName(SboSalesOrderTable.NAME, SboSalesOrderTable.Columns.DocEntry.NAME));
                {
                    strBuilder.AppendFormat("\r\n    SELECT {0}",
                        SqlCommandHelper.EscapeTableColumnName(SboSalesDeliveryNoteLineTable.NAME, SboDocumentLineColumns.BaseEntry.NAME));

                    strBuilder.AppendFormat("\r\n    FROM {0}",
                        SqlCommandHelper.EscapeTableName(SboSalesDeliveryNoteLineTable.NAME));

                    strBuilder.AppendFormat("\r\n    WHERE {0} = {1}",
                        SqlCommandHelper.EscapeTableColumnName(SboSalesDeliveryNoteLineTable.NAME, SboSalesDeliveryNoteLineTable.Columns.DocEntry.NAME),
                        SqlCommandHelper.FormatAndEscapeColumnValue(salesDeliveryDocEntry, dbCulture));
                    strBuilder.AppendFormat(" AND {0} = {1}",
                        SqlCommandHelper.EscapeTableColumnName(SboSalesDeliveryNoteLineTable.NAME, SboDocumentLineColumns.BaseType.NAME),
                        SqlCommandHelper.FormatAndEscapeColumnValue(((int)SboObjectType.SalesOrder).ToString(), dbCulture));
                }
                strBuilder.Append("\r\n)");

                strBuilder.AppendFormat("\r\nGROUP BY {0}",
                    SqlCommandHelper.EscapeTableColumnName(SboSalesOrderTable.NAME, SboSalesOrderTable.Columns.DocTotal.NAME));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(SboSalesOrderTable.NAME, SboSalesOrderTable.Columns.DocTotalFC.NAME));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(SboSalesOrderTable.NAME, SboSalesOrderTable.Columns.DocCurrency.NAME));

                return strBuilder.ToString();
            }

            /// <summary>
            /// Gets all LUID's for sales order.
            /// </summary>
            /// <param name="salesOrderDocEntry">The sales order doc entry.</param>
            /// <param name="dbCulture">The db culture.</param>
            /// <returns></returns>
            internal static string GetAllLUIDsForSalesDelivery(int salesDeliveryDocEntry, CultureInfo dbCulture)
            {
                StringBuilder strBuilder = new StringBuilder();

                strBuilder.AppendFormat("SELECT {0}",
                    SqlCommandHelper.EscapeTableColumnName(PmxLogisticUnitIDTable.NAME, PmxLogisticUnitIDTable.Columns.InternalKey.NAME));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(PmxLogisticUnitIDTable.NAME, PmxLogisticUnitIDTable.Columns.SSCC.NAME));

                strBuilder.AppendFormat("\r\nFROM {0}",
                    SqlCommandHelper.EscapeTableName(PmxLinkSalesDeliveryNoteLinePickListLineTable.NAME));

                strBuilder.AppendFormat("\r\nINNER JOIN {0}",
                    SqlCommandHelper.EscapeTableName(PmxMoveLineTable.NAME));
                strBuilder.AppendFormat(" ON {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName(PmxMoveLineTable.NAME, PmxMoveLineTable.Columns.DocEntry.NAME),
                    SqlCommandHelper.EscapeTableColumnName(PmxLinkSalesDeliveryNoteLinePickListLineTable.NAME, PmxLinkSalesDeliveryNoteLinePickListLineTable.Columns.MoveLineDocEntry.NAME));
                strBuilder.AppendFormat(" AND {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName(PmxMoveLineTable.NAME, PmxMoveLineTable.Columns.LineNum.NAME),
                    SqlCommandHelper.EscapeTableColumnName(PmxLinkSalesDeliveryNoteLinePickListLineTable.NAME, PmxLinkSalesDeliveryNoteLinePickListLineTable.Columns.MoveLineLineNum.NAME));

                strBuilder.AppendFormat("\r\nLEFT JOIN {0}",
                    SqlCommandHelper.EscapeTableName(PmxLogisticUnitIDTable.NAME));
                strBuilder.AppendFormat(" ON {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName(PmxLogisticUnitIDTable.NAME, PmxLogisticUnitIDTable.Columns.InternalKey.NAME),
                    SqlCommandHelper.EscapeTableColumnName(PmxMoveLineTable.NAME, PmxMoveLineTable.Columns.DestinationLogisticUnitIdentificationKey.NAME));

                strBuilder.AppendFormat("\r\nWHERE {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName(PmxLinkSalesDeliveryNoteLinePickListLineTable.NAME, PmxLinkSalesDeliveryNoteLinePickListLineTable.Columns.SalesDeliveryNoteLineDocEntry.NAME),
                    SqlCommandHelper.FormatAndEscapeColumnValue(salesDeliveryDocEntry, dbCulture));

                strBuilder.AppendFormat("\r\nGROUP BY {0}",
                    SqlCommandHelper.EscapeTableColumnName(PmxLogisticUnitIDTable.NAME, PmxLogisticUnitIDTable.Columns.InternalKey.NAME));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnName(PmxLogisticUnitIDTable.NAME, PmxLogisticUnitIDTable.Columns.SSCC.NAME));

                return strBuilder.ToString();
            }

            /// <summary>
            /// Updates the LUID.
            /// </summary>
            /// <param name="logUnitIdentKey">The log unit ident key.</param>
            /// <param name="trackingNumber">The tracking number.</param>
            /// <param name="dbCulture">The db culture.</param>
            /// <returns></returns>
            internal static string UpdateLUID(int logUnitIdentKey, string trackingNumber, CultureInfo dbCulture)
            {
                StringBuilder strBuilder = new StringBuilder();

                strBuilder.AppendFormat("UPDATE {0}",
                    SqlCommandHelper.EscapeTableName(PmxLogisticUnitIDTable.NAME));

                strBuilder.AppendFormat("\r\nSET {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName(PmxLogisticUnitIDTable.NAME, PmxLogisticUnitIDTableLenL.Columns.TrnspTrackID.NAME),
                    SqlCommandHelper.FormatAndEscapeColumnValue(trackingNumber, dbCulture));

                strBuilder.AppendFormat("\r\nWHERE {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName(PmxLogisticUnitIDTable.NAME, PmxLogisticUnitIDTable.Columns.InternalKey.NAME),
                    SqlCommandHelper.FormatAndEscapeColumnValue(logUnitIdentKey, dbCulture));

                return strBuilder.ToString();
            }
        }
    }
}